adler32() {
	perl -e '
		$sum1=1;
		$sum2=0;
		while ($bytesRead = read (STDIN, $buffer, 1)) {
			$val = unpack 'C', $buffer;
			$sum1 = ($sum1 + $val) % 65521;
			$sum2 = ($sum2 + $sum1) % 65521;
		}
		$adler = $sum1 + 65536 * $sum2;
		printf("%08x", $adler & 0xffffffff);
	'
}

oldworldchecksum() {
	# Ben Boldt's Mac ROM Checksum Verifier
	# https://web.archive.org/web/20180217090913/https://www.d.umn.edu/~bold0070/projects/checksum/
	perl -e '
		$ck = 0;
		while ($bytesRead = read (STDIN, $buffer, 2)) {
			$num = unpack 'n', $buffer;
			$ck += $num;
		} 
		printf("%08x", $ck & 0xffffffff);
	'
}

machine_name() {
	local config_info_offset=$((0x$(dd if="$therom" bs=1 skip=0x300080 count=4 2> /dev/null | xxd -p)))
	local rom_id_str=$(dd if="$therom" bs=1 skip=$((0x300064 + $config_info_offset)) count=16 2> /dev/null)
	printf "\"%s\"" "${rom_id_str}" | perl -pe 's/[^ -~]/./g'
}

doonerom() {
	local therom="$1"
	echo $(
		romsize=$(cat "$therom" | wc -c)
		thebytes=$(dd if="$therom" bs=32 count=1 2> /dev/null | xxd -p)
		theversion=$((0x${thebytes:8:8}))
		theyear=$((0x${thebytes:16:4}))
		if (($romsize == 1048576)) && ((theyear > 0x1990 && theyear < 0x2006)); then
			productcode=""
			if ((theyear < 0x2000)); then
				calcedchecksum1="$(perl -0777 -ne 'print substr($_,0,0x3ffc)' "$therom" | adler32)"
				actualchecksum1="$(perl -0777 -ne 'print substr($_,0x3ffc,4)' "$therom" | xxd -p)"
				calcedchecksum2="00000000";
				actualchecksum2="00000000";
				configchecksum="";
			else
				calcedchecksum1="$(perl -0777 -ne 'print substr($_,0,0x3efc)' "$therom" | adler32)"
				actualchecksum1="$(perl -0777 -ne 'print substr($_,0x3efc,4)' "$therom" | xxd -p)"
				calcedchecksum2="$(perl -0777 -ne 'print substr($_,0x3f00,0x7c)' "$therom" | adler32)"
				configchecksum="$( perl -0777 -ne 'print substr($_,0x3f0c,0x70)' "$therom" | adler32)"
				actualchecksum2="$(perl -0777 -ne 'print substr($_,0x3f7c,4)' "$therom" | xxd -p)"
				productcode="0x$(perl -0777 -ne 'print substr($_,0x3f02,2); print substr($_,0x3f13,1)' "$therom" | xxd -p)"
			fi
			calcedchecksum3="$(perl -0777 -ne 'print substr($_,0x8000,0x77ffc)' "$therom" | adler32)"
			actualchecksum3="$(perl -0777 -ne 'print substr($_,0x7fffc,4)' "$therom" | xxd -p)"
			calcedchecksum4="$(perl -0777 -ne 'print substr($_,0x80000,0x7fffc)' "$therom" | adler32)"
			actualchecksum4="$(perl -0777 -ne 'print substr($_,0xffffc,4)' "$therom" | xxd -p)"
			printf "%s.%s %-5x %-18s %4d %s " "${thebytes:16:4}" "${thebytes:20:4}" \
				"$theversion" "${productcode:0:6}.${productcode:6:2} ${configchecksum}" $((romsize / 1024)) \
				"${actualchecksum1}.${actualchecksum2}.${actualchecksum3}.${actualchecksum4}"
			[[ "${actualchecksum1}.${actualchecksum2}.${actualchecksum3}.${actualchecksum4}" == \
			   "${calcedchecksum1}.${calcedchecksum2}.${calcedchecksum3}.${calcedchecksum4}" ]] && \
				printf "               √" || \
				printf  "${calcedchecksum1}.${calcedchecksum2}.${calcedchecksum3}.${calcedchecksum4} x"
		else
			checksum="$(perl -0777 -ne 'print substr($_,4,0x2ffffc)' "$therom" | oldworldchecksum)"
			printf "%s.%s %-18s %4d %s " "${thebytes:16:4}" "${thebytes:36:4}" \
				"$(machine_name)" $((romsize / 1024)) "${thebytes:0:8}"
			[[ $checksum == ${thebytes:0:8} ]] && printf "         √ " || printf  "${checksum} x "
			md5 -q "$therom"
		fi
	) '"'"$therom"'"'
}

doallroms() {
	IFS=$'\n'
	for therom in $(find . \( -size 64k -o -size 128k -o -size 256k -o -size 512k -o -size 1M -o -size 2M -o -size 3M -o -size 4M -o -size 5M -o -size 6M \) \
		\( -iname '*.rom' -o -name 'MacROM' \)  \
		-not -path "*/ROM ATI*" \
		-not -path "*/ROM Nvidia*" \
		-not -path "*/ROM Firmtek*" \
		-not -path "*/ROM Adaptec*" \
		-not -name "@FFF80000 len-80000 Rom Image.rom" \
	); do
		echo "$therom" 1>&2
		doonerom "$therom"
	done | sort
}
